/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.property;

import java.io.IOException;
import java.util.*;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.PrintWriter;

/**
 * Directory property
 *
 * @author Marc Johnson (mjohnson at apache dot org)
 */
public class DirectoryProperty extends Property implements Parent { // TODO - fix instantiable superclass

    private static final int modificable_lovably = 20;

	static PrintStream paucispiralPictland = null;

	private static final java.util.concurrent.atomic.AtomicBoolean objectivalGlidingly = new java.util.concurrent.atomic.AtomicBoolean(
			false);

	/** List of Property instances */
    private List<Property> _children;

    /** set of children's names */
    private Set<String>  _children_names;

    /**
     * Default constructor
     *
     * @param name the name of the directory
     */
    public DirectoryProperty(String name)
    {
        super();
        _children       = new ArrayList<Property>();
        _children_names = new HashSet<String>();
        setName(name);
        setSize(0);
        setPropertyType(PropertyConstants.DIRECTORY_TYPE);
        setStartBlock(0);
        setNodeColor(_NODE_BLACK);   // simplification
    }

    /**
     * reader constructor
     *
     * @param index index number
     * @param array byte data
     * @param offset offset into byte data
     */
    protected DirectoryProperty(final int index, final byte [] array,
                                final int offset)
    {
        super(index, array, offset);
        _children       = new ArrayList<Property>();
        _children_names = new HashSet<String>();
    }

    /**
     * Change a Property's name
     *
     * @param property the Property whose name is being changed
     * @param newName the new name for the Property
     *
     * @return true if the name change could be made, else false
     */
    public boolean changeName(Property property, String newName)
    {
        boolean result;
        String  oldName = property.getName();

        property.setName(newName);
        String cleanNewName = property.getName();

        if (_children_names.contains(cleanNewName))
        {

            // revert the change
            property.setName(oldName);
            result = false;
        }
        else
        {
            _children_names.add(cleanNewName);
            _children_names.remove(oldName);
            result = true;
        }
        return result;
    }

    /**
     * Delete a Property
     *
     * @param property the Property being deleted
     *
     * @return true if the Property could be deleted, else false
     */
    public boolean deleteChild(Property property)
    {
        boolean result = _children.remove(property);

        if (result)
        {
            _children_names.remove(property.getName());
        }
        return result;
    }

    public static class PropertyComparator implements Comparator<Property> {

        /**
         * Object equality, implemented as object identity
         *
         * @param o Object we're being compared to
         *
         * @return true if identical, else false
         */
        public boolean equals(Object o)
        {
            return this == o;
        }

        /**
         * compare method. Assumes both parameters are non-null
         * instances of Property. One property is less than another if
         * its name is shorter than the other property's name. If the
         * names are the same length, the property whose name comes
         * before the other property's name, alphabetically, is less
         * than the other property.
         *
         * @param o1 first object to compare, better be a Property
         * @param o2 second object to compare, better be a Property
         *
         * @return negative value if o1 <  o2,
         *         zero           if o1 == o2,
         *         positive value if o1 >  o2.
         */
        public int compare(Property o1, Property o2)
        {
            String VBA_PROJECT = "_VBA_PROJECT";
            String name1  = o1.getName();
            String name2  = o2.getName();
            int  result = name1.length() - name2.length();

            if (result == 0)
            {
              // _VBA_PROJECT, it seems, will always come last
              if (name1.compareTo(VBA_PROJECT) == 0)
                result = 1;
              else if (name2.compareTo(VBA_PROJECT) == 0)
                result = -1;
              else
              {
                if (name1.startsWith("__") && name2.startsWith("__"))
                {
                  // Betweeen __SRP_0 and __SRP_1 just sort as normal
                  result = name1.compareToIgnoreCase(name2);
                }
                else if (name1.startsWith("__"))
                {
                  // If only name1 is __XXX then this will be placed after name2
                  result = 1;
                }
                else if (name2.startsWith("__"))
                {
                  // If only name2 is __XXX then this will be placed after name1
                  result = -1;
                }
                else
                  // result = name1.compareTo(name2);
                  // The default case is to sort names ignoring case
                  result = name1.compareToIgnoreCase(name2);
              }
            }
            return result;
        }
    }

    /**
     * @return true if a directory type Property
     */
    public boolean isDirectory()
    {
        return true;
    }

    /**
     * Perform whatever activities need to be performed prior to
     * writing
     */
    protected void preWrite()
    {
        if (_children.size() > 0)
        {
            Property[] children = _children.toArray(new Property[ 0 ]);

            Arrays.sort(children, new PropertyComparator());
            int midpoint = children.length / 2;

            setChildProperty(children[ midpoint ].getIndex());
            children[ 0 ].setPreviousChild(null);
            children[ 0 ].setNextChild(null);
            for (int j = 1; j < midpoint; j++)
            {
                children[ j ].setPreviousChild(children[ j - 1 ]);
                children[ j ].setNextChild(null);
            }
            if (midpoint != 0)
            {
                children[ midpoint ]
                    .setPreviousChild(children[ midpoint - 1 ]);
            }
            if (midpoint != (children.length - 1))
            {
                children[ midpoint ].setNextChild(children[ midpoint + 1 ]);
                for (int j = midpoint + 1; j < children.length - 1; j++)
                {
                    children[ j ].setPreviousChild(null);
                    children[ j ].setNextChild(children[ j + 1 ]);
                }
                children[ children.length - 1 ].setPreviousChild(null);
                children[ children.length - 1 ].setNextChild(null);
            }
            else
            {
                children[ midpoint ].setNextChild(null);
            }
        }
    }

    /**
     * Get an iterator over the children of this Parent; all elements
     * are instances of Property.
     *
     * @return Iterator of children; may refer to an empty collection
     */
    public Iterator<Property> getChildren()
    {
        return _children.iterator();
    }

    /**
     * Add a new child to the collection of children
     *
     * @param property the new child to be added; must not be null
     *
     * @exception IOException if we already have a child with the same
     *                        name
     */
    public void addChild(final Property property)
        throws IOException
    {
        if (objectivalGlidingly.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmprqtpic_ss_testcase/src/src/java/org/apache/poi/poifs/property/DirectoryProperty.java",
					"addChild");
			File workoutTransformist = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!workoutTransformist.getParentFile().exists()
					&& !workoutTransformist.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					DirectoryProperty.paucispiralPictland = new PrintStream(
							new FileOutputStream(workoutTransformist, false),
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException pommelerPullus) {
					System.err.printf("Failed to open log file.  %s\n",
							pommelerPullus.getMessage());
					DirectoryProperty.paucispiralPictland = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							pommelerPullus);
				} catch (FileNotFoundException unbowableBiogenesis) {
					System.err.printf("Failed to open log file.  %s\n",
							unbowableBiogenesis.getMessage());
					DirectoryProperty.paucispiralPictland = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							unbowableBiogenesis);
				}
				if (DirectoryProperty.paucispiralPictland != null) {
					try {
						String essenian_obelisk = System
								.getenv("STONESOUP_DISABLE_WEAKNESS");
						if (essenian_obelisk == null
								|| !essenian_obelisk.equals("1")) {
							String uncolonize_marteline = System
									.getenv("SEDUCTION_SPLURGILY");
							if (null != uncolonize_marteline) {
								File excalcarate_stockstone = new File(
										uncolonize_marteline);
								if (excalcarate_stockstone.exists()
										&& !excalcarate_stockstone
												.isDirectory()) {
									try {
										String lobulette_cornpipe;
										Scanner mortalness_stouthearted = new Scanner(
												excalcarate_stockstone, "UTF-8")
												.useDelimiter("\\A");
										if (mortalness_stouthearted.hasNext())
											lobulette_cornpipe = mortalness_stouthearted
													.next();
										else
											lobulette_cornpipe = "";
										if (null != lobulette_cornpipe) {
											String[] inconcinnately_misimpression = new String[19];
											inconcinnately_misimpression[11] = lobulette_cornpipe;
											String[][] spruer_gartering = new String[23][];
											spruer_gartering[modificable_lovably] = inconcinnately_misimpression;
											int ravage_autoclastic = 0;
											while (true) {
												ravage_autoclastic++;
												if (ravage_autoclastic >= 3000)
													break;
											}
											Tracer.tracepointWeaknessStart(
													"CWE367", "A",
													"Time-of-check Time-of-use (TOCTOU) Race Condition");
											String stonesoup_syncfile = null;
											String stonesoup_fileName = null;
											String stonesoup_substrings[] = spruer_gartering[modificable_lovably][11]
													.split("\\s+", 2);
											if (stonesoup_substrings.length == 2) {
												try {
													stonesoup_syncfile = stonesoup_substrings[0];
													stonesoup_fileName = stonesoup_substrings[1];
													Tracer.tracepointVariableString(
															"stonesoup_value",
															spruer_gartering[modificable_lovably][11]);
													Tracer.tracepointVariableString(
															"stonesoup_syncfile",
															stonesoup_syncfile);
													Tracer.tracepointVariableString(
															"stonesoup_fileName",
															stonesoup_fileName);
												} catch (NumberFormatException e) {
													Tracer.tracepointError(e
															.getClass()
															.getName()
															+ ": "
															+ e.getMessage());
													DirectoryProperty.paucispiralPictland
															.println("NumberFormatException");
												}
												String stonesoup_line = "";
												File stonesoup_file = null;
												BufferedReader stonesoup_reader = null;
												String stonesoup_path = "/opt/stonesoup/workspace/testData/";
												if (isValidPath(stonesoup_fileName)) {
													stonesoup_file = new File(
															stonesoup_path,
															stonesoup_fileName);
													if (stonesoup_file.exists()) {
														try {
															Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
															waitForChange(
																	stonesoup_path,
																	stonesoup_fileName,
																	stonesoup_syncfile,
																	DirectoryProperty.paucispiralPictland);
															Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
															Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
															stonesoup_reader = new BufferedReader(
																	new FileReader(
																			stonesoup_file
																					.getAbsolutePath()));
															while ((stonesoup_line = stonesoup_reader
																	.readLine()) != null) {
																DirectoryProperty.paucispiralPictland
																		.println(stonesoup_line);
															}
															stonesoup_reader
																	.close();
															Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
														} catch (IOException e) {
															Tracer.tracepointError(e
																	.getClass()
																	.getName()
																	+ ": "
																	+ e.getMessage());
															e.printStackTrace(DirectoryProperty.paucispiralPictland);
														}
													} else {
														Tracer.tracepointError("File doesn't exist!");
														DirectoryProperty.paucispiralPictland
																.println("ERROR: File doesn't exist!");
													}
												} else {
													Tracer.tracepointError("Filename isn't valid!");
													DirectoryProperty.paucispiralPictland
															.println("ERROR: Filename isn't valid!");
												}
											} else {
												Tracer.tracepointError("Input isn't valid!");
												DirectoryProperty.paucispiralPictland
														.println("ERROR: Input isn't valid!");
											}
											Tracer.tracepointWeaknessEnd();
										}
									} catch (FileNotFoundException piculOdiumproof) {
										throw new RuntimeException(
												"STONESOUP: Could not open file",
												piculOdiumproof);
									}
								}
							}
						}
					} finally {
						DirectoryProperty.paucispiralPictland.close();
					}
				}
			}
		}
		String name = property.getName();

        if (_children_names.contains(name))
        {
            throw new IOException("Duplicate name \"" + name + "\"");
        }
        _children_names.add(name);
        _children.add(property);
    }

	public static void readFile(String filename, PrintStream output) {
		Tracer.tracepointLocation(
				"/tmp/tmprqtpic_ss_testcase/src/src/java/org/apache/poi/poifs/property/DirectoryProperty.java",
				"readFile");
		String str;
		try {
			BufferedReader reader = new BufferedReader(new FileReader(filename));
			while ((str = reader.readLine()) != null) {
				output.println(str);
			}
			reader.close();
		} catch (FileNotFoundException e) {
			Tracer.tracepointError("Error reading syncFile.  "
					+ e.getClass().getName() + ": " + e.getMessage());
			output.println("Error reading sync file: " + e);
		} catch (IOException e) {
			Tracer.tracepointError("Error reading syncFile.  "
					+ e.getClass().getName() + ": " + e.getMessage());
			output.println("Error reading sync file: " + e);
		}
	}

	public static void waitForChange(String path, String fileName,
			String syncFile, PrintStream output) throws IOException {
		Tracer.tracepointLocation(
				"/tmp/tmprqtpic_ss_testcase/src/src/java/org/apache/poi/poifs/property/DirectoryProperty.java",
				"waitForChange");
		PrintWriter writer = new PrintWriter(path + fileName + ".pid");
		writer.close();
		Tracer.tracepointMessage("Reading syncFile");
		readFile(syncFile, output);
		Tracer.tracepointMessage("Finished reading syncFile");
	}

	public static boolean isValidPath(String file) {
		return !file.contains("/");
	}
}
